//#include "String.mqh"
#include "Ticks.mqh"

// https://www.mql5.com/ru/forum/170952/page125#comment_10619096
// #define PRINT(A) ::Print(A + " (Memory used = " + (string)::TerminalInfoInteger(TERMINAL_MEMORY_USED) + " MB)")
#define PRINT(A) ::Print(A)

class CUSTOMSYMBOL : public SYMBOL
{
private:
  TICKS Ticks;

public:
  CUSTOMSYMBOL( const string Symb = NULL, const string Path = NULL, const string SymbOriginal = NULL ) : SYMBOL(Symb, Path, SymbOriginal)
  {
  }

  bool SwapTicks( MqlTick &TicksArray[] )
  {
    return(this.Ticks.Swap(TicksArray));
  }

  uint AddTicks( const MqlTick &TicksArray[] )
  {
    return(this.Ticks.Add(TicksArray));
  }

  uint AddTicks( const string &Str[] )
  {
    return(this.Ticks.Add(Str));
  }

  uint AddTicks( const uchar &Bytes[], const uint StartPos = 0, const int Reserve = 0 )
  {
/*    string Str[];

    STRING::BytesToStrings(Bytes, Str);

    return(this.AddTicks(Str)); */

    this.Ticks.SetReserve(Reserve);

    return(this.Ticks.Add(Bytes, StartPos));
  }

  uint AddTicks( const string FileName, const uint StartPos = 0  )
  {
    uchar Bytes[];

    ::FileLoad(FileName, Bytes);

    return(this.AddTicks(Bytes, StartPos));
  }

  void CreateSessions( void ) const
  {
    datetime From[];
    datetime To[];

    for (int i = this.Ticks.GetSessions(From, To) - 1; i >= 0; i--)
      if (From[i] < To[i])
      {
        datetime FromQuote;
        datetime ToQuote;

        if (::SymbolInfoSessionQuote(this.Name, (ENUM_DAY_OF_WEEK)i, 0, FromQuote, ToQuote))
        {
          From[i] = ::MathMin(From[i], FromQuote);

          To[i] = ::MathMax(To[i], ToQuote);
        }

        ::CustomSymbolSetSessionQuote(this.Name, (ENUM_DAY_OF_WEEK)i, 0, From[i], To[i]);

        //        - BestInterval.

        datetime FromTrade;
        datetime ToTrade;

        if (::SymbolInfoSessionTrade(this.Name, (ENUM_DAY_OF_WEEK)i, 0, FromTrade, ToTrade))
        {
          From[i] = ::MathMin(From[i], FromTrade);

          To[i] = ::MathMax(To[i], ToTrade);
        }

        ::CustomSymbolSetSessionTrade(this.Name, (ENUM_DAY_OF_WEEK)i, 0, From[i], To[i]);
      }

    return;
  }

  int CreateHistoryTicks( const double TickValue = 1 )
  {
    PRINT("Calculating Digits...");
    this.SetProperty(SYMBOL_DIGITS, this.Ticks.GetDigits());

//    this.SetProperty(SYMBOL_TRADE_CONTRACT_SIZE, ::MathPow(10, this.GetProperty(SYMBOL_DIGITS)));
    this.SetProperty(SYMBOL_TRADE_CONTRACT_SIZE, 1e5); // https://www.mql5.com/ru/forum/330333#comment_14608694
    this.SetProperty(SYMBOL_TRADE_TICK_VALUE, 1);
/*
    this.SetProperty(SYMBOL_TRADE_TICK_VALUE, ::SymbolInfoDouble(_Symbol, SYMBOL_TRADE_TICK_VALUE));
    this.SetProperty(SYMBOL_TRADE_TICK_VALUE_LOSS, this.GetProperty(SYMBOL_TRADE_TICK_VALUE));
    this.SetProperty(SYMBOL_TRADE_TICK_VALUE_PROFIT, this.GetProperty(SYMBOL_TRADE_TICK_VALUE));
*/
    this.SetProperty(SYMBOL_TRADE_TICK_SIZE, this.GetProperty(SYMBOL_POINT));

    const string Currency = ::AccountInfoString(ACCOUNT_CURRENCY);

    this.SetProperty(SYMBOL_CURRENCY_PROFIT, Currency);
    this.SetProperty(SYMBOL_CURRENCY_MARGIN, Currency);
    this.SetProperty(SYMBOL_CURRENCY_BASE, Currency);

    // https://www.mql5.com/ru/forum/321656/page34#comment_14192799
    PRINT("Normalizing Ticks...");
    this.Ticks.Normalize((int)this.GetProperty(SYMBOL_DIGITS));

    PRINT("Correcting Ticks...");
    const int Errors = this.Ticks.Correct();

    if (Errors)
      PRINT("Corrected " + (string)Errors + " ticks.");

    PRINT("Calculating Sessions...");
    this.CreateSessions();

    const uint Amount = this.Ticks.GetAmount();
    PRINT("Saving Ticks: " + (Amount ? (string)this.Ticks[0].time + " - " + (string)this.Ticks[Amount - 1].time : "NULL."));

    const ulong StartTime = ::GetMicrosecondCount();
    const int Res = this.Ticks.ToSymbol(&this);
    const ulong StopTime = ::GetMicrosecondCount();

    PRINT("Saved ticks = " + (string)Res + ((StopTime == StartTime) ? NULL :
            " (" + (string)((ulong)Res * 1000000 / (StopTime - StartTime)) + " ticks/sec)"));

    return(Res);
  }

  int CreateHistoryRates( void ) const
  {
    RATES Rates((int)this.GetProperty(SYMBOL_DIGITS));

    PRINT("Generating Rates...");
    PRINT("Generated Rates = " + (string)this.Ticks.ToRates(&Rates));

    const uint Amount = Rates.GetAmount();
    PRINT("Saving Rates: " + (Amount ? (string)Rates[0].time + " - " + (string)Rates[Amount - 1].time : "NULL."));

    const ulong StartTime = ::GetMicrosecondCount();
    const int Res = Rates.ToSymbol(&this);
    const ulong StopTime = ::GetMicrosecondCount();

    PRINT("Saved Rates = " + (string)Res + ((StopTime == StartTime) ? NULL :
            " (" + (string)((ulong)Res * 1000000 / (StopTime - StartTime)) + " rates/sec)"));

    return(Res);
  }

  //       BarInterval.
  int DataToSymbol( const int MinPips = -1, const int MinInterval = 7 * 24 * 3600, const int BarInterval = 0, const bool AntiDST = false )
  {
    if (MinPips >= 0)
    {
      //    
      // https://www.mql5.com/ru/forum/212096/page2#comment_7018318
      this.SetProperty(SYMBOL_TRADE_EXEMODE, SYMBOL_TRADE_EXECUTION_EXCHANGE);
      this.SetProperty(SYMBOL_TRADE_CALC_MODE, SYMBOL_CALC_MODE_EXCH_FUTURES);

      const uint BeforeFilter = this.Ticks.GetAmount();
      const uint AfterFilter = this.Ticks.FilterTicks(MinPips, BarInterval);

      PRINT("After Filter (MinPips = " + (string)MinPips + ", BarInterval = " + (string)BarInterval + ") Ticks = " + (string)AfterFilter +
              (BeforeFilter ? " (" + ::DoubleToString(100.0 * AfterFilter / BeforeFilter, 2) + "%)" : NULL));
      PRINT("Add " + (string)this.Ticks.AddLastPrices() + " last prices.");

      if (AntiDST)
      {
        ::Print("AntiDST...");

        this.Ticks.AntiDST();
      }
    }

    // https://www.mql5.com/ru/forum/360210/page43#comment_20793287
    PRINT("Add " + (string)this.Ticks.AddFirstTicks(MinInterval) + " first ticks.");

    const int Res = this.CreateHistoryTicks();

    this.CreateHistoryRates();

    return(Res);
  }
};

#undef PRINT